 package receiver;

/*
 *	MidiCommon.java
 *
 *	This file is part of jsresources.org
 */

/*
 * Copyright (c) 1999 - 2001 by Matthias Pfisterer
 * Copyright (c) 2003 by Florian Bomers
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
|<---            this code is formatted to fit into 80 columns             --->|
*/

import java.util.ArrayList;
import java.util.Vector;

import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.ShortMessage;



/** 
 * This is a utility class for obtaining MIDI devices connected to the computer.
 * @author Ki-Hyun. Sung
 * @author Jason. Park
 * @author Edmund. Zurita
 * */
public class MidiCommon
{
	
	private static final String[] sm_astrKeyNames = 
		{"C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"};
	
	/**	
	 * This method shows the input and/or output MIDI devices connected to the computer. 
	 * @param bForInput - shows input devices if true
	 * 
	 * @param bForOutput - shows output devices if true 
	 * 
	 * @return Vector devices
	 */
	public static Vector listDevicesAndExit(boolean bForInput,
					      boolean bForOutput)
	{
		return listDevicesAndExit(bForInput, bForOutput, false);
	}


	/**	
	 * This method shows the input and/or output MIDI devices connected to the computer. 
	 * Can specify to return version information.
	 * @param bForInput - shows input devices if true
	 * 
	 * @param bForOutput - shows output devices if true
	 * 
	 * @param bVerbose - shows version information of the devices if true
	 *  
	 * @return A vector containing the the names of the devices found
	 */
	public static Vector listDevicesAndExit(boolean bForInput,
					      boolean bForOutput,
					      boolean bVerbose)
	{
		
		
		if (bForInput && !bForOutput)
		{
			out("Available MIDI IN Devices:");
		}
		else if (!bForInput && bForOutput)
		{
			out("Available MIDI OUT Devices:");
		}
		else
		{
			out("Available MIDI Devices:");
		}
		Vector<String> devices = new Vector<String>();
		//devices.add("Select MIDI input device");
		MidiDevice.Info[]	aInfos = MidiSystem.getMidiDeviceInfo();
		devices.add("Choose a device");
		devices.add("#  Keyboard Simulator");
		for (int i = 0; i < aInfos.length; i++)
		{
			try
			{
				MidiDevice	device = MidiSystem.getMidiDevice(aInfos[i]);
				boolean		bAllowsInput = (device.getMaxTransmitters() != 0);
				boolean		bAllowsOutput = (device.getMaxReceivers() != 0);
				if ((bAllowsInput && bForInput) ||
				    (bAllowsOutput && bForOutput))
				{
					if (bVerbose)
					{
					out("" + i + "  "
						+ (bAllowsInput?"IN ":"   ")
						+ (bAllowsOutput?"OUT ":"    ")
						+ aInfos[i].getName() + ", "
						+ aInfos[i].getVendor() + ", "
						+ aInfos[i].getVersion() + ", "
						+ aInfos[i].getDescription());
					}
					else
					{
						
						devices.add("" + i + "  " + aInfos[i].getName());
						out("" + i + "  " + aInfos[i].getName());
					}
				}
			}
			catch (MidiUnavailableException e)
			{
				// device is obviously not available...
				// out(e);
			}
		}
		if (aInfos.length == 0)
		{
			out("[No devices available]");
		}		
		return devices;
		//System.exit(0);
	}



	/** Retrieve a MidiDevice.Info for a given name.

	This method tries to return a MidiDevice.Info whose name
	matches the passed name. If no matching MidiDevice.Info is
	found, null is returned.  If bForOutput is true, then only
	output devices are searched, otherwise only input devices.

	@param strDeviceName - the name of the device for which an info
	object should be retrieved.

	@param bForOutput - If true, only output devices are
	considered. If false, only input devices are considered.

	@return A MidiDevice.Info object matching the passed device
	name or null if none could be found.

	*/
	public static MidiDevice.Info getMidiDeviceInfo(String strDeviceName, boolean bForOutput)
	{
		MidiDevice.Info[]	aInfos = MidiSystem.getMidiDeviceInfo();
		for (int i = 0; i < aInfos.length; i++)
		{
			if (aInfos[i].getName().equals(strDeviceName))
			{
				try
				{
					MidiDevice device = MidiSystem.getMidiDevice(aInfos[i]);
					boolean	bAllowsInput = (device.getMaxTransmitters() != 0);
					boolean	bAllowsOutput = (device.getMaxReceivers() != 0);
					if ((bAllowsOutput && bForOutput) || (bAllowsInput && !bForOutput))
					{
						return aInfos[i];
					}
				}
				catch (MidiUnavailableException e)
				{
					// TODO:
				}
			}
		}
		return null;
	}


	/** 
	 * Retrieve a MidiDevice.Info by index number.
	 * This method returns a MidiDevice.Info whose index
	 * is specified as parameter. This index matches the
	 * number printed in the listDevicesAndExit method.
	 * If index is too small or too big, null is returned.
	 *
	 * @param index - the index of the device to be retrieved
	 * @return A MidiDevice.Info object of the specified index
	 *         or null if none could be found.
	 */
	public static MidiDevice.Info getMidiDeviceInfo(int index)
	{
		MidiDevice.Info[]	aInfos = MidiSystem.getMidiDeviceInfo();
		if ((index < 0) || (index >= aInfos.length)) {
			return null;
		}
		return aInfos[index];
	}

	private static void out(String strMessage)
	{
		System.out.println(strMessage);
	}
	
	/**
	 * Finds out whether a sequence of message is a chord or a single note.
	 * 
	 * @param m - an ArrayList containing a sequence of ShortMessages. 
	 * 
	 * @param n - an ArrayList containing a sequence of timestamps.
	 * 
	 * @return A string representation of the chord or the note recognized from the sequence.
	 */
	public static String recognizeChord(ArrayList m){
		String returnStr = new String();

		for(int i=0; i<m.size(); i++){

			ShortMessage msg;
			ShortMessage msg1;
			
			if(i!=0){
				msg = (ShortMessage)(m.get(i));
				msg1 = (ShortMessage)(m.get(i-1));
			}
			else{
				msg = (ShortMessage)(m.get(i));
				msg1 = (ShortMessage)(m.get(i));
			}
					
			if(msg.getStatus()==144 && msg1.getStatus()!=128){
				returnStr+=getKeyName(msg.getData1());
			}
			else if(msg.getStatus()==144){
				returnStr+=getKeyName(msg.getData1());
			}
		}
		
		//checkForPopularChord(returnStr);
		return returnStr;
		
	}
	
	public static String recognizeChordForSoundSimulator(ArrayList m){
		String returnStr = new String();

		for(int i=0; i<m.size(); i++){

			ShortMessage msg;
			ShortMessage msg1;
			
			if(i!=0){
				msg = (ShortMessage)(m.get(i));
				msg1 = (ShortMessage)(m.get(i-1));
			}
			else{
				msg = (ShortMessage)(m.get(i));
				msg1 = (ShortMessage)(m.get(i));
			}
					
			if(msg.getStatus()==144 && msg1.getStatus()!=128){
				returnStr+=getKeyNameForSoundSimulator(msg.getData1());
			}
			else if(msg.getStatus()==144){
				returnStr+=getKeyNameForSoundSimulator(msg.getData1());
			}
		}
		
		//checkForPopularChord(returnStr);
		return returnStr;
		
	}
	
	/**
	 * Gets the Key name given the MIDI integer representation of the key
	 * @param nKeyNumber - the MIDI integer representation of the key
	 * @return The string representation of the key.
	 */
	public static String getKeyName(int nKeyNumber)
	{
		if (nKeyNumber > 127)
		{
			return "illegal value";
		}
		else
		{
			int	nNote = nKeyNumber % 12;
			int	nOctave = nKeyNumber / 12;
			return sm_astrKeyNames[nNote] + (nOctave - 1);
		}
	}
	
	public static String getKeyNameForSoundSimulator(int nKeyNumber)
	{
		if (nKeyNumber > 127)
		{
			return "illegal value";
		}
		else
		{
			int	nNote = nKeyNumber % 12;
			int	nOctave = nKeyNumber / 12;
			if(nNote == 9 && nOctave == 6){
				return "H5";
			}
			return sm_astrKeyNames[nNote] + (nOctave - 1);
		}
	}
	
	/**
	 * 
	 * @param message - A ShortMessage received by the send method.
	 * @return - A string representation of the ShortMessage. 
	 * 			 Includes timestamp, command, key and velocity information.s
	 */
	public static String decodeMessage(ShortMessage message)
	{
		String	strMessage = null;
		switch (message.getCommand())
		{
		case 0x80:
			
			strMessage = "note Off " + getKeyName(message.getData1()) + " velocity: " + message.getData2();
			break;

		case 0x90:
			strMessage = "note On " + getKeyName(message.getData1()) + " velocity: " + message.getData2();
			break;

		case 0xa0:
			strMessage = "polyphonic key pressure " + getKeyName(message.getData1()) + " pressure: " + message.getData2();
			break;

		case 0xb0:
			strMessage = "control change " + message.getData1() + " value: " + message.getData2();
			break;

		case 0xc0:
			strMessage = "program change " + message.getData1();
			break;

		case 0xd0:
			strMessage = "key pressure " + getKeyName(message.getData1()) + " pressure: " + message.getData2();
			break;

		case 0xe0:
			strMessage = "pitch wheel change " + get14bitValue(message.getData1(), message.getData2());
			break;

/*		case 0xF0:
			strMessage = SYSTEM_MESSAGE_TEXT[message.getChannel()];
			switch (message.getChannel())
			{
			case 0x1:
				int	nQType = (message.getData1() & 0x70) >> 4;
				int	nQData = message.getData1() & 0x0F;
				if (nQType == 7)
				{
					nQData = nQData & 0x1;
				}
				strMessage += QUARTER_FRAME_MESSAGE_TEXT[nQType] + nQData;
				if (nQType == 7)
				{
					int	nFrameType = (message.getData1() & 0x06) >> 1;
					strMessage += ", frame type: " + FRAME_TYPE_TEXT[nFrameType];
				}
				break;

			case 0x2:
				strMessage += get14bitValue(message.getData1(), message.getData2());
				break;

			case 0x3:
				strMessage += message.getData1();
				break;
			}
			break;*/

		default:
			strMessage = "unknown message: status = " + message.getStatus() + ", byte1 = " + message.getData1() + ", byte2 = " + message.getData2();
			break;
		}
		
		//System.out.println(getHexString(message));
		if (message.getCommand() != 0xF0)
		{
			int	nChannel = message.getChannel() + 1;
			String	strChannel = "channel " + nChannel + ": ";
			strMessage = strChannel + strMessage;
		}
		//smCount++;
		//smByteCount+=message.getLength();
		
		
		
		return "["+getHexString(message)+"] "+strMessage;
	}

	/**
	 * Gets the Key name given the MIDI integer representation of the key
	 * @param nKeyNumber - the MIDI integer representation of the key
	 * @return The string representation of the key.
	 */

	public static int get14bitValue(int nLowerPart, int nHigherPart)
	{
		return (nLowerPart & 0x7F) | ((nHigherPart & 0x7F) << 7);
	}

	/**
	 * Gets the Key name given the MIDI integer representation of the key
	 * @param nKeyNumber - the MIDI integer representation of the key
	 * @return The string representation of the key.
	 */

/*	private static int signedByteToUnsigned(byte b)
	{
		return b & 0xFF;
	}
	*//**
	 * Gets the Key name given the MIDI integer representation of the key
	 * @param nKeyNumber - the MIDI integer representation of the key
	 * @return The string representation of the key.
	 *//*
	// convert from microseconds per quarter note to beats per minute and vice versa
	private static float convertTempo(float value) {
		if (value <= 0) {
			value = 0.1f;
		}
		return 60000000.0f / value;
	}*/



	private static char hexDigits[] = 
	   {'0', '1', '2', '3', 
	    '4', '5', '6', '7', 
	    '8', '9', 'A', 'B', 
	    'C', 'D', 'E', 'F'};
	/**
	 * Gets the Hexadecimal representation of a byte array
	 * @param sm - The byte array
	 * @return The string representation.
	 */
	public static String getHexString(byte[] aByte)
	{
		StringBuffer	sbuf = new StringBuffer(aByte.length * 3 + 2);
		for (int i = 0; i < aByte.length; i++)
		{
			sbuf.append(' ');
			sbuf.append(hexDigits[(aByte[i] & 0xF0) >> 4]);
			sbuf.append(hexDigits[aByte[i] & 0x0F]);
			/*byte	bhigh = (byte) ((aByte[i] &  0xf0) >> 4);
			sbuf.append((char) (bhigh > 9 ? bhigh + 'A' - 10: bhigh + '0'));
			byte	blow = (byte) (aByte[i] & 0x0f);
			sbuf.append((char) (blow > 9 ? blow + 'A' - 10: blow + '0'));*/
		}
		return new String(sbuf);
	}
	/**
	 * Converts integer to hexadecimal.
	 * @param i - the integer number.
	 * @return The string representation.
	 */
	private static String intToHex(int i) {
		return ""+hexDigits[(i & 0xF0) >> 4]
		         +hexDigits[i & 0x0F];
	}
	/**
	 * Gets the Hexadecimal representation of the ShortMessage
	 * @param sm - The ShortMessage
	 * @return The string representation.
	 */
	public static String getHexString(ShortMessage sm)
	{
		// bug in J2SDK 1.4.1
		// return getHexString(sm.getMessage());
		int status = sm.getStatus();
		String res = intToHex(sm.getStatus());
		// if one-byte message, return
		switch (status) {
			case 0xF6:			// Tune Request
			case 0xF7:			// EOX
	    		// System real-time messages
			case 0xF8:			// Timing Clock
			case 0xF9:			// Undefined
			case 0xFA:			// Start
			case 0xFB:			// Continue
			case 0xFC:			// Stop
			case 0xFD:			// Undefined
			case 0xFE:			// Active Sensing
			case 0xFF: return res;
		}
		res += ' '+intToHex(sm.getData1());
		// if 2-byte message, return
		switch (status) {
			case 0xF1:			// MTC Quarter Frame
			case 0xF3:			// Song Select
					return res;
		}
		switch (sm.getCommand()) {
			case 0xC0:
			case 0xD0:
					return res;
		}
		// 3-byte messages left
		res += ' '+intToHex(sm.getData2());
		return res;
	}
}



/*** MidiCommon.java ***/


